package org.xparse;

import java.io.IOException;
import java.io.InputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.AccessibleObject;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class XParse {
	private static final Logger log = Logger.getLogger(XParse.class);
	// private XPath xPath;
	private Document xmlDocument;

	public XParse() {
		// xPath = XPathFactory.newInstance().newXPath();
	}

	public XParse(URL url) {
		this();
		setDocument(url);
	}

	public void setDocument(Document doc) {
		xmlDocument = doc;
	}

	public void setDocument(InputStream is) {
		try {

			Document doc = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder().parse(is);
			setDocument(doc);
		} catch (IOException ex) {
			throw new RuntimeException("Can't open XML", ex);
		} catch (SAXException ex) {
			throw new RuntimeException("Can't parse XML", ex);
		} catch (ParserConfigurationException ex) {
			throw new RuntimeException("Cant configure parser", ex);
		}
	}

	public void setDocument(URL url) {
		try {
			InputStream is = url.openStream();
			setDocument(is);
			is.close();
		} catch (IOException ex) {
			throw new RuntimeException("Can't open XML", ex);
		}
	}

	public Object parse(Object obj, String xpathstr) {
		try {
			XPath xpath = new DOMXPath(xpathstr);
			Node node = (Node) xpath.selectSingleNode(xmlDocument);
			cache.process(obj, node);
			return obj;
		} catch (IllegalArgumentException ex) {
			throw new RuntimeException("Can't call method of given object", ex);
		} catch (DOMException ex) {
			throw new RuntimeException("Can't parese DOM", ex);
		} catch (JaxenException ex) {
			throw new RuntimeException("Jaxen error", ex);
		}
	}
	
	/**
	 * Process given XML file, using given xpath and creates objects of class E
	 * based on the structure of XML. Nodes will be processed recursively
	 * XParseAttr or XParseValue.
	 * 
	 * @param <E>
	 * @param class_
	 * @param url
	 * @return list of objects found by xpath taken from XParseNode annotation
	 */
	public <E> List<E> parse(Class<E> class_, String xpathstr) {
		try {
			XPath xpath = new DOMXPath(xpathstr);
			List<Node> nodes = xpath.selectNodes(xmlDocument);
			List<E> result = new LinkedList<E>();
			for (Node node : nodes) {
				E e = class_.newInstance();
				cache.process(e, node);
				result.add(e);
			}
			return result;
		} catch (InstantiationException ex) {
			throw new RuntimeException("Problem instantiating given class", ex);
		} catch (IllegalAccessException ex) {
			throw new RuntimeException("Can't access given class", ex);
		} catch (IllegalArgumentException ex) {
			throw new RuntimeException("Can't call method of given object", ex);
		} catch (DOMException ex) {
			throw new RuntimeException("Can't parese DOM", ex);
		} catch (JaxenException ex) {
			throw new RuntimeException("Jaxen error", ex);
		}
	}

	private List<String> getNodeValue(Node node) {
		List<String> texts = new LinkedList<String>();

		if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
			texts.add(node.getNodeValue());
		} else {
			NodeList children = node.getChildNodes();

			for (int i = 0; i < children.getLength(); i++) {
				Node child = children.item(i);
				if (child.getNodeName().equals("#text")) {
					texts.add(child.getNodeValue());
				}
			}
		}

		return texts;
	}

	private List<String> compactList(List<String> texts) {
		List<String> textsCompact = new LinkedList<String>();

		for (String str : texts) {
			String compacted = str.trim();
			if (!compacted.equals("")) {
				textsCompact.add(compacted);
			}
		}

		return textsCompact;
	}

	Cache cache = new Cache();

	class Cache {
		Map<Class, List<Executor>> cache = new HashMap<Class, List<Executor>>();

		public Object process(Object obj, Node node) {
			Class class_ = obj.getClass();
			List<Executor> executors = cache.get(class_);
			if (executors == null) {
				executors = buildExecutors(class_);
				cache.put(class_, executors);
			}

			for (Executor executor : executors) {
				executor.execute(obj, node);
			}

			return obj;
		}

		private Executor createExecutor(AnnotatedElement ae){
			XParseValue lavann = ae.getAnnotation(XParseValue.class);
			if (lavann != null) {
				return new ValueExecutor(ae, lavann);
			}

			XParseAttr attrann = ae.getAnnotation(XParseAttr.class);
			if (attrann != null) {
				return new AttributeExecutor(ae, attrann);
			}

			XParseNode nodeann = ae.getAnnotation(XParseNode.class);
			if (nodeann != null) {
				Class class_ = nodeann.clazz();
				if (class_.equals(XParse.class)){
					if (ae instanceof Field) {
						if (nodeann.collection()){
							class_ = String.class;
						} else {
							class_ = ((Field) ae).getType();
						}
					} else {
						class_ = ((Method) ae).getParameterTypes()[0];
					}
				}
				boolean stringMethod = class_.equals(String.class);
				if (stringMethod) {
					return new StringNodeExecutor(ae, nodeann);
				} else {
					return new ComplexTypeExecutor(ae, nodeann);
				}
			} else {
				return null;
			}
		}
		
		public List<Executor> buildExecutors(Class class_) {
			List<Executor> result = new LinkedList<Executor>();
			for (Method method : class_.getDeclaredMethods()) {
				Executor e = createExecutor(method);
				if (e!=null){
					result.add(e);
				}
			}
			for (Field field : class_.getDeclaredFields()) {
				Executor e = createExecutor(field);
				if (e != null){
					result.add(e);
				}
			}
			return result;
		}
	}

	abstract class Executor {
		AnnotatedElement ae;
		Annotation ann;
		XPath exp;

		public Executor(AnnotatedElement method, Annotation ann) {
			this.ae = method;
			this.ann = ann;
		}

		public void execute(Object obj, Node node) {
			try {
				_execute(obj, node);
			} catch (JaxenException ex) {
				throw new RuntimeException("Wrong XPath", ex);
			} catch (IllegalAccessException ex) {
				throw new RuntimeException("Can't access given class", ex);
			} catch (IllegalArgumentException ex) {
				throw new RuntimeException("Can't call method of given object",
						ex);
			} catch (DOMException ex) {
				throw new RuntimeException("Can't parese DOM", ex);
			} catch (InvocationTargetException ex) {
				throw new RuntimeException("Can't call method of given object",
						ex);
			} catch (InstantiationException e) {
				throw new RuntimeException("Can't instantiate object", e);
			}

		}

		public abstract void _execute(Object obj, Node node)
				throws IllegalArgumentException, DOMException,
				IllegalAccessException, InvocationTargetException,
				InstantiationException, JaxenException;
	}

	class AttributeExecutor extends Executor {
		String attr;

		public AttributeExecutor(AnnotatedElement ae, XParseAttr ann) {
			super(ae, ann);
			attr = ann.name();
		}

		@Override
		public void _execute(Object obj, Node node)
				throws IllegalArgumentException, DOMException,
				IllegalAccessException, InvocationTargetException {
			Node attrNode = node.getAttributes().getNamedItem(attr);
			if (attrNode != null) {
				((AccessibleObject)ae).setAccessible(true);
				if (ae instanceof Field) {
					((Field)ae).set(obj, attrNode.getNodeValue());
				} else {
					((Method)ae).invoke(obj, attrNode.getNodeValue());
				}
			}
		}
	}

	class ValueExecutor extends Executor {
		public ValueExecutor(AnnotatedElement ae, XParseValue ann) {
			super(ae, ann);
		}

		@Override
		public void _execute(Object obj, Node node)
				throws IllegalArgumentException, IllegalAccessException,
				InvocationTargetException {
			List<String> texts = getNodeValue(node);
			List<String> textsCompact = compactList(texts);

			String fullText = StringUtils.join(texts, "");
			String fullTextCompact = StringUtils.join(textsCompact, "");
			((AccessibleObject)ae).setAccessible(true);
			if (ae instanceof Method){
				if (((Method)ae).getParameterTypes()[0].equals(String.class)) {
					((Method)ae).invoke(obj,
							((XParseValue) ann).compact() ? fullTextCompact
									: fullText);
				} else {
					((Method)ae).invoke(obj, ((XParseValue) ann).compact() ? textsCompact
							: texts);
				}
			} else {
				Field field = (Field)ae;
				if (field.getType().equals(String.class)){
					field.set(obj,
						((XParseValue) ann).compact() ? fullTextCompact
								: fullText);
				} else {
					field.set(obj, ((XParseValue) ann).compact() ? textsCompact
							: texts);
				}
			}
		}

	}

	abstract class XPathExecutor extends Executor {
		XPath exp;
		public XPathExecutor(AnnotatedElement ae, XParseNode ann) {
			super(ae, ann);
			try {
				exp = new DOMXPath(ann.xpath());
			} catch (JaxenException e) {
				throw new RuntimeException("Can't compile xpath: "
						+ ann.xpath(), e);
			}
		}

	}

	class StringNodeExecutor extends XPathExecutor {
		public StringNodeExecutor(AnnotatedElement ae, XParseNode ann) {
			super(ae, ann);
		}

		@Override
		public void _execute(Object obj, Node node)
				throws IllegalArgumentException, DOMException,
				IllegalAccessException, InvocationTargetException,
				JaxenException {
			Iterable<Node> nodes = exp.selectNodes(node);
			for (Node innerNode : nodes) {
				String str = StringUtils.join(getNodeValue(innerNode), "");
				((AccessibleObject)ae).setAccessible(true);
				if (ae instanceof Field) {
					boolean isCollection = ((Field) ae).get(obj) instanceof Collection;
					if (isCollection){
						((Collection)((Field) ae).get(obj)).add(str);
					} else {
						((Field)ae).set(obj, str);	
					}
				} else {
					((Method)ae).invoke(obj, str);
				}
			}
			//"".get
		}
		
		

	}

	class ComplexTypeExecutor extends XPathExecutor {
		public ComplexTypeExecutor(AnnotatedElement ae, XParseNode ann) {
			super(ae, ann);
		}

		@Override
		public void _execute(Object obj, Node node)
				throws IllegalArgumentException, DOMException,
				IllegalAccessException, InvocationTargetException,
				InstantiationException, JaxenException {

			Iterable<Node> nodes = exp.selectNodes(node);
			for (Node innerNode : nodes) {
				Class class_ = ((XParseNode)ann).clazz();
				Object innerobj = null;
				if (!class_.equals(XParse.class)){
					innerobj = class_.newInstance();
				}
				((AccessibleObject)ae).setAccessible(true);
				if (ae instanceof Field) {
					if (((XParseNode)ann).collection()){
						assert innerobj != null;
						cache.process(innerobj, innerNode);
						((Collection)((Field) ae).get(obj)).add(innerobj);
					} else {
						innerobj = ((Field)ae).getDeclaringClass().newInstance();
						cache.process(innerobj, innerNode);
						((Field)ae).set(obj, innerobj);
					}
				} else {
					assert innerobj==null;
					innerobj = ((Method)ae).getParameterTypes()[0].newInstance();
					if (((XParseNode)ann).collection()){
						throw new RuntimeException("XParseNode.collection() cannot be used with methods");
					}
					cache.process(innerobj, innerNode);
					((Method)ae).invoke(obj, innerobj);
				}
				
				
			}
		}
	}

}